# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.project import Project

# Python3 compat
try:
    unicode = str
except NameError:  # pragma: no cover
    pass


@click.command('init')
@click.pass_context
@click.option('-s', '--scons', is_flag=True,
              help='Create default SConstruct file.')
@click.option('-b', '--board', type=unicode, metavar='BOARD',
              help='Create init file with the selected board.')
@click.option('--project-dir', type=unicode, metavar='PATH',
              help='Set the target directory for the project')
def cli(ctx, board, scons, project_dir):
    """Manage apio projects."""

    if scons:
        Project().create_sconstruct(project_dir)
    elif board:
        Project().create_ini(board, project_dir)
    else:
        click.secho(ctx.get_help())



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import re
import click
import platform

from os.path import join, isdir

from apio import util


class System(object):  # pragma: no cover
    def __init__(self):
        self.ext = ''
        if 'Windows' == platform.system():
            self.ext = '.exe'

    def lsusb(self):
        self._run('listdevs')

    def lsftdi(self):
        self._run('find_all')

    def detect_boards(self):
        detected_boards = []
        result = self._run('find_all')

        if result and result['returncode'] == 0:
            detected_boards = self.parse_out(result['out'])

        return detected_boards

    def _run(self, command):
        result = []
        system_dir = join(util.get_home_dir(), 'packages', 'system')

        if isdir(system_dir):
            result = util.exec_command(
                join(system_dir, command + self.ext),
                stdout=util.AsyncPipe(self._on_run_out),
                stderr=util.AsyncPipe(self._on_run_out)
                )
        else:
            click.secho('Error: system tools are not installed', fg='red')
            click.secho('Please run:\n'
                        '   apio install system', fg='yellow')

        return result

    def _on_run_out(self, line):
        click.secho(line)

    def parse_out(self, text):
        pattern = 'Number\sof\sFTDI\sdevices\sfound:\s(?P<n>\d+?)\n'
        match = re.search(pattern, text)
        n = int(match.group('n')) if match else 0

        pattern = '.*Checking\sdevice:\s(?P<index>.*?)\n.*'
        index = re.findall(pattern, text)

        pattern = '.*Manufacturer:\s(?P<n>.*?),.*'
        manufacturer = re.findall(pattern, text)

        pattern = '.*Description:\s(?P<n>.*?)\n.*'
        description = re.findall(pattern, text)

        detected_boards = []

        for i in range(n):
            board = {
                "index": index[i],
                "manufacturer": manufacturer[i],
                "description": description[i]
            }
            detected_boards.append(board)

        return detected_boards



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import json
from os.path import isfile, join

from apio.util import get_home_dir


class Profile(object):

    def __init__(self):
        self.packages = {}
        self._profile_path = join(get_home_dir(), 'profile.json')
        self.load()

    def check_package(self, name):
        return (name in self.packages.keys())

    def check_version(self, name, version):
        return not ((name in self.packages.keys()) and
                    (self.packages[name]['version'] >= version))

    def add(self, name, version):
        self.packages[name] = {'version': version}

    def remove(self, name):
        if name in self.packages.keys():
            del self.packages[name]

    def get_version(self, name):
        return self.packages[name]['version']

    def load(self):
        self.packages = {}
        if isfile(self._profile_path):
            with open(self._profile_path, 'r') as profile:
                try:
                    self.packages = json.load(profile)
                except:
                    pass
                profile.close()

    def save(self):
        with open(self._profile_path, 'w') as profile:
            json.dump(self.packages, profile)
            profile.close()



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.installer import Installer
from apio.resources import Resources


@click.command('uninstall')
@click.pass_context
@click.argument('packages', nargs=-1)
@click.option('-a', '--all', is_flag=True,
              help='Uninstall all packages.')
@click.option('-l', '--list', is_flag=True,
              help='List all installed packages.')
def cli(ctx, packages, all, list):
    """Uninstall packages."""

    if packages:
        _uninstall(packages)
    elif all:  # pragma: no cover
        packages = Resources().packages
        _uninstall(packages)
    elif list:
        Resources().list_packages(installed=True, notinstalled=False)
    else:
        click.secho(ctx.get_help())


def _uninstall(packages):
    if click.confirm('Do you want to continue?'):
        for package in packages:
            if package == 'pio-fpga':  # skip pio-fpga
                continue
            Installer(package).uninstall()
    else:
        click.secho('Abort!', fg='red')



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2
# -- Derived from:
# ---- Platformio project
# ---- (C) 2014-2016 Ivan Kravets <me@ikravets.com>
# ---- Licence Apache v2

import os
import subprocess
from os.path import expanduser, isdir, join
from platform import system, uname
from threading import Thread

import requests
requests.packages.urllib3.disable_warnings()

__version__ = None

# Python3 compat
try:
    unicode = str
except NameError:  # pragma: no cover
    pass


class ApioException(Exception):

    MESSAGE = None

    def __str__(self):  # pragma: no cover
        if self.MESSAGE:
            return self.MESSAGE.format(*self.args)
        else:
            return Exception.__str__(self)


class AbortedByUser(ApioException):

    MESSAGE = "Aborted by user"


class AsyncPipe(Thread):  # pragma: no cover

    def __init__(self, outcallback=None):
        Thread.__init__(self)
        self.outcallback = outcallback

        self._fd_read, self._fd_write = os.pipe()
        self._pipe_reader = os.fdopen(self._fd_read)
        self._buffer = []

        self.start()

    def get_buffer(self):
        return self._buffer

    def fileno(self):
        return self._fd_write

    def run(self):
        for line in iter(self._pipe_reader.readline, ""):
            line = line.strip()
            self._buffer.append(line)
            if self.outcallback:
                self.outcallback(line)
            else:
                print(line)
        self._pipe_reader.close()

    def close(self):
        os.close(self._fd_write)
        self.join()


def get_systype():
    data = uname()
    arch = ''
    type_ = data[0].lower()
    if type_ == 'linux':
        arch = data[4].lower() if data[4] else ""
    return "%s_%s" % (type_, arch) if arch else type_


def _get_projconf_option_dir(name, default=None):
    _env_name = "APIO_%s" % name.upper()
    if _env_name in os.environ:
        return os.getenv(_env_name)
    return default


def get_home_dir():
    home_dir = _get_projconf_option_dir(
        "home_dir",
        join(expanduser("~"), ".apio")
    )

    if not isdir(home_dir):
        os.makedirs(home_dir)

    assert isdir(home_dir)
    return home_dir


def get_project_dir():
    return os.getcwd()


def change_filemtime(path, time):
    os.utime(path, (time, time))


def is_ci():
    return os.getenv("CI", "").lower() == "true"


def exec_command(*args, **kwargs):  # pragma: no cover
    result = {
        "out": None,
        "err": None,
        "returncode": None
    }

    default = dict(
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=system() == "Windows"
    )
    default.update(kwargs)
    kwargs = default

    p = subprocess.Popen(*args, **kwargs)
    try:
        result['out'], result['err'] = p.communicate()
        result['returncode'] = p.returncode
    except KeyboardInterrupt:
        raise AbortedByUser()
    finally:
        for s in ("stdout", "stderr"):
            if isinstance(kwargs[s], AsyncPipe):
                kwargs[s].close()

    for s in ("stdout", "stderr"):
        if isinstance(kwargs[s], AsyncPipe):
            result[s[3:]] = "\n".join(kwargs[s].get_buffer())

    for k, v in result.items():
        if v and isinstance(v, unicode):
            result[k].strip()

    return result


def get_request_defheaders():
    return {"User-Agent": "PlatformIO/%s CI/%d %s" % (
        __version__, int(is_ci()), requests.utils.default_user_agent()
    )}


def get_pypi_latest_version():
    r = requests.get("https://pypi.python.org/pypi/apio/json",
                     headers=get_request_defheaders())
    r.raise_for_status()
    return r.json()['info']['version']



from apio import cli as cmd_apio


def test_apio(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_apio)
    validate_cliresult(result)


def test_apio_wrong_command(clirunner, validate_cliresult):
    result = clirunner.invoke(cmd_apio, ['missing_command'])
    assert result.exit_code == 2
    assert 'Error: No such command' in result.output



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click
import subprocess

from os.path import join, dirname, isfile

from apio.util import get_systype

platform = get_systype()


class Drivers(object):  # pragma: no cover

    rules_local_path = join(
        dirname(__file__), '..', 'resources', '80-icestick.rules')
    rules_system_path = '/etc/udev/rules.d/80-icestick.rules'

    def enable(self):
        if 'linux' in platform:
            self._enable_linux()
        elif 'darwin' in platform:
            self._enable_darwin()
        elif 'windows' in platform:
            self._enable_windows()

    def disable(self):
        if 'linux' in platform:
            self._disable_linux()
        elif 'darwin' in platform:
            self._disable_darwin()
        elif 'windows' in platform:
            self._disable_windows()

    def _enable_linux(self):
        click.secho('Configure FTDI drivers for FPGA')
        if not isfile(self.rules_system_path):
            subprocess.call(['sudo', 'cp',
                             self.rules_local_path, self.rules_system_path])
            subprocess.call(['sudo', 'service', 'udev', 'restart'])
            click.secho('FPGA drivers enabled', fg='green')
        else:
            click.secho('Already enabled', fg='yellow')

    def _disable_linux(self):
        if isfile(self.rules_system_path):
            click.secho('Revert FTDI drivers\' configuration')
            subprocess.call(['sudo', 'rm', self.rules_system_path])
            click.secho('FPGA drivers disabled', fg='yellow')
        else:
            click.secho('Already disabled', fg='red')

    def _enable_darwin(self):
        # TODO: return if brew is not installed
        subprocess.call(['brew', 'install', 'libftdi'])
        click.secho('Configure FTDI drivers for FPGA')
        subprocess.call(['sudo', 'kextunload', '-b',
                         'com.FTDI.driver.FTDIUSBSerialDriver'])
        subprocess.call(['sudo', 'kextunload', '-b',
                         'com.apple.driver.AppleUSBFTDI'])
        click.secho('FPGA drivers enabled', fg='green')

    def _disable_darwin(self):
        click.secho('Revert FTDI drivers\' configuration')
        subprocess.call(['sudo', 'kextload', '-b',
                         'com.FTDI.driver.FTDIUSBSerialDriver'])
        subprocess.call(['sudo', 'kextload', '-b',
                         'com.apple.driver.AppleUSBFTDI'])
        click.secho('FPGA drivers disabled', fg='green')

    def _enable_windows(self):
        import webbrowser
        url = 'https://github.com/FPGAwars/apio/wiki/enableation#windows'
        click.secho('Follow the next instructions: ' + url)
        webbrowser.open(url)

    def _disable_windows(self):
        pass



# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='apio',
    version='0.1.4',
    author='Jesús Arroyo Torrens',
    author_email='jesus.jkhlg@gmail.com',
    url='https://github.com/FPGAwars/apio',
    license='GPLv2',
    packages=['apio'],
    package_data={
        'apio': ['commands/*.py',
                 'managers/*.py',
                 'resources/*']
    },
    install_requires=[
        'click',
        'requests'
    ],
    entry_points={
        'console_scripts': ['apio=apio:cli']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python']
)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Juan González, Jesús Arroyo
# -- Licence GPLv2

import sys
import json
import click

from os.path import isfile, join, dirname

from apio.resources import Resources

PROJECT_FILENAME = 'apio.ini'


class Project(object):

    def __init__(self):
        self.board = None

    def create_sconstruct(self, project_dir=''):
        """Creates a default SConstruct file"""

        if project_dir is None:
            project_dir = ''

        sconstruct_name = 'SConstruct'
        sconstruct_path = join(project_dir, sconstruct_name)
        local_sconstruct_path = join(
            dirname(__file__), '..', 'resources', sconstruct_name)

        if isfile(sconstruct_path):
            click.secho(
                'Warning: {} file already exists'.format(sconstruct_name),
                fg='yellow')
            if click.confirm('Do you want to replace it?'):
                self._copy_sconstruct_file(sconstruct_name, sconstruct_path,
                                           local_sconstruct_path)
            else:
                click.secho('Abort!', fg='red')
        else:
            self._copy_sconstruct_file(sconstruct_name, sconstruct_path,
                                       local_sconstruct_path)

    def create_ini(self, board, project_dir=''):
        """Creates a new apio project file"""

        if project_dir is None:
            project_dir = ''

        ini_path = join(project_dir, PROJECT_FILENAME)

        # Check board
        boards = Resources().boards
        if board not in boards.keys():
            click.secho(
                'Error: No such board \'{}\''.format(board),
                fg='red')
            sys.exit(1)

        if isfile(ini_path):
            click.secho(
                'Warning: {} file already exists'.format(PROJECT_FILENAME),
                fg='yellow')
            if click.confirm('Do you want to replace it?'):
                self._create_ini_file(board, ini_path, PROJECT_FILENAME)
            else:
                click.secho('Abort!', fg='red')
        else:
            self._create_ini_file(board, ini_path, PROJECT_FILENAME)

    def _create_ini_file(self, board, ini_path, ini_name):
        # Creates the project dictionary
        project = {"board": board}

        # Dump the project into the apio.ini file
        project_str = json.dumps(project)

        click.secho('Creating {} file ...'.format(ini_name))
        with open(ini_path, 'w') as file:
            file.write(project_str)
            click.secho(
                'File \'{}\' has been successfully created!'.format(
                    ini_name),
                fg='green')

    def _copy_sconstruct_file(self, sconstruct_name,
                              sconstruct_path, local_sconstruct_path):
        click.secho('Creating {} file ...'.format(sconstruct_name))
        with open(sconstruct_path, 'w') as sconstruct:
            with open(local_sconstruct_path, 'r') as local_sconstruct:
                sconstruct.write(local_sconstruct.read())
                click.secho(
                    'File \'{}\' has been successfully created!'.format(
                        sconstruct_name),
                    fg='green')

    def read(self):
        """Read the project config file"""

        # -- If no project finel found, just return
        if not isfile(PROJECT_FILENAME):
            print("Info: No apio.ini file")
            return

        # -- Open the project file
        with open(PROJECT_FILENAME, 'r') as f:
            project_str = f.read()

        # -- Decode the jsonj
        try:
            project = json.loads(project_str)
        except:
            print("Error: Invalid {} project file".format(PROJECT_FILENAME))
            sys.exit(1)

        # -- TODO: error checking

        # -- Update the board
        try:
            self.board = project['board']
        except:
            print("Error: Invalid {} project file".format(PROJECT_FILENAME))
            print("No 'board' field defined in project file")
            sys.exit(1)



# -*- coding: utf-8 -*-
# -- This file is part of the Apio project
# -- (C) 2016 FPGAwars
# -- Author Jesús Arroyo
# -- Licence GPLv2

import click

from apio.managers.scons import SCons


@click.command('verify')
@click.pass_context
def cli(ctx):
    """Verify the verilog code."""
    exit_code = SCons().verify()
    ctx.exit(exit_code)
